use std::fmt;

use crate::Span;

/// Kinds of errors which can occur during lexical analysis.
#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum LexerErrorKind {
    /// An invalid character literal was encountered.
    InvalidChar,
    /// An invalid keyword was encountered.
    InvalidKeyword,
    /// An invalid number literal was encountered.
    InvalidNumber,
    /// An invalid symbol was encountered.
    InvalidSymbol,
    /// An unclosed string literal was encountered.
    UnclosedString,
}

#[cfg(not(tarpaulin_include))]
impl fmt::Display for LexerErrorKind {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        use LexerErrorKind::*;

        match self {
            InvalidChar => write!(f, "Invalid character literal"),
            InvalidKeyword => write!(f, "Invalid keyword"),
            InvalidNumber => write!(f, "Invalid number literal"),
            InvalidSymbol => write!(f, "Invalid symbol"),
            UnclosedString => write!(f, "Unclosed string literal"),
        }
    }
}

/// Errors which occur during lexical analysis.
#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct LexerError {
    /// The kind of lexer error.
    pub kind: LexerErrorKind,
    /// The span of the lexer error.
    pub span: Span,
    /// Additional context regarding the lexer error.
    pub context: Option<String>,
}

impl LexerError {
    /// Construct a new instance of a lexer error.
    #[must_use]
    pub const fn new(kind: LexerErrorKind, span: Span) -> Self {
        Self {
            kind,
            span,
            context: None,
        }
    }

    /// Provide additional context for a lexer error.
    #[must_use]
    pub fn with_context<C>(mut self, f: impl FnOnce() -> C) -> Self
    where
        C: fmt::Display,
    {
        self.context = Some(f().to_string());
        self
    }
}

impl std::error::Error for LexerError {}

#[cfg(not(tarpaulin_include))]
impl fmt::Display for LexerError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        if let Some(ref context) = self.context {
            write!(f, "{}: {}", self.kind, context)
        } else {
            write!(f, "{}", self.kind)
        }
    }
}
