use crate::Span;

/// Kinds of tokens which are valid in Onihime source code.
#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum TokenKind {
    /// Line comment, e.g. `; ...`
    Comment,
    /// Whitespace, e.g. ' ', '\t', '\n'
    Whitespace,

    /// Opening parenthesis, e.g. `(`
    OpenParen,
    /// Closing parenthesis, e.g. `)`
    CloseParen,
    /// Opening brace, e.g. `{`
    OpenBrace,
    /// Closing brace, e.g. `}`
    CloseBrace,
    /// Opening bracket, e.g. `[`
    OpenBracket,
    /// Closing bracket, e.g. `]`
    CloseBracket,

    /// Boolean, e.g. `true`, `false`
    Bool,
    /// Character, e.g. `\a`, `\x1e`, `\u03BB`, `\newline`
    Char,
    /// Keyword, e.g. `:foo-bar`, `:baz`, `:qux0`
    Keyword,
    /// Floating-point number, e.g. `-1.0`, `2.0`, `3.0e-4`
    Decimal,
    /// Integer, e.g. `0`, `-1`, `0b1010`, `0o7`, `0xDECAFBAD`
    Integer,
    /// Ratio, e.g. `1/3`, `-5/7`
    Ratio,
    /// String, e.g. `"foo bar"`
    String,
    /// Symbol, e.g. `baz`, `*qux*`, `nil?`, `+`
    Symbol,
    /// Nil, e.g. `nil`
    Nil,
}

impl TokenKind {
    /// Returns `true` if the token type an atom.
    pub fn is_atom(&self) -> bool {
        use TokenKind::*;

        matches!(
            self,
            Bool | Char | Keyword | Decimal | Integer | Ratio | String | Symbol | Nil
        )
    }
}

/// A valid token found in Onihime source code.
#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub struct Token {
    /// Kind of token which was found.
    pub kind: TokenKind,
    /// The token's span.
    pub span: Span,
}

impl Token {
    /// Construct a new instance of a token.
    #[must_use]
    pub const fn new(kind: TokenKind, span: Span) -> Self {
        Self { kind, span }
    }
}
